from typing import TYPE_CHECKING, Any

from pydantic import BaseModel

from pyromax.types import Message
from pyromax.types.OpcodeEnum import Opcode


class Chat(BaseModel):
    max_api: Any
    id: int = None
    owner: int
    participants: dict
    status: str
    type: str
    options: dict | None = None
    restrictions: int | None = None
    has_bots: str | None = None
    access: str | None = None
    last_message: Message | Message = None

    @classmethod
    def from_json(cls, json, max_api: 'MaxApi') -> list['Chat']:

        if not isinstance(json, list):
            json = [json]
        chats = []
        for json_chat in json:
            instance = cls(**json_chat, max_api=max_api)
            if 'lastMessage' in json_chat:
                instance.last_message = Message(**json_chat['lastMessage'], max_api=max_api, chatId=instance.id, opcode=Opcode.SEND_MESSAGE.value)

            chats.append(instance)
        return chats


    def __repr__(self):
        details = []
        for key, value in self.__dict__.items():
            details.append(f"{key}={value}")
        return '\n'.join(details)


    async def get_messages_per_chunk(self, time: int) -> list[Message]:
        response = await self.max_api.max_client.send_and_receive(opcode=Opcode.GET_CHAT_MESSAGES_PER_CHUNK, payload={
            'chatId': self.id,
            'from': time,
            'forward': 0,
            'backward': 30,
            'getMessages': True,

        })

        response = response[0]
        messages = []

        for message in response['payload']['messages']:
            messages.append(Message(**message, max_api=self.max_api, chatId=self.id))

        return messages

    async def get_all_messages(self, time: int) -> list[Message]:


        # pprint(response)
        # while response['opcode'] == 128:
        #     response = await self.max_client.wait_recv()
        #     response = response[0]
        #     print('============')
        #     pprint(response)

        messages = await self.get_messages_per_chunk(time)
        chunk = await self.get_messages_per_chunk(time)
        messages = chunk + messages

        while len(chunk) >= 30:
            chunk = await self.get_messages_per_chunk(chunk[0].time)
            messages = chunk + messages

        return messages


import asyncio
from asyncio import Future
from socket import gaierror
import logging
from typing import Iterable

from websockets import WebSocketException

# from .ObserverPattern import Subject
from .Router import Router
from pyromax.types import Update, Message, Opcode
from pyromax.api import MaxApi
from pyromax.utils import NotFoundFlag
from pyromax.utils import get_dict_value_by_path


class Dispatcher(Router):
    def __init__(self):
        super().__init__()

        self.__logger = logging.getLogger('MaxDispatcher')


    # async def notify(self, update, data: dict):
    #     if not update:
    #         return
    #     result = await Router.notify(self, update, data)
    #     if result:
    #         return
    #     for router in self.sub_routers:
    #         for sub_router in router.chain_head:
    #             result = await router.notify(update=update, data=data)
    #             if result:
    #                 break


    async def _check_update(self, max_api: MaxApi):
        update = await max_api.max_client.wait_recv(cmd=0)
        if update:
            update = Update(**(update[0]), max_api=max_api)
            data = {
                Opcode: update.opcode,
                Update: update,
            }
            data.update(max_api.base_data)


            self.__logger.debug(f'Dispatcher update: %s', update)
            for event in self.events.values():
                if event.opcode == update.opcode:

                    parsed_update = event.type_of_update.from_update(update)
                    data = parsed_update.edit_data(data)
                    return parsed_update, data
        self.__logger.debug('Dispatcher update skipped: %s', update)
        return NotFoundFlag(), NotFoundFlag()


    async def start_polling(self, max_api: MaxApi):
        while True:
            if max_api.max_client:
                update, data = await self._check_update(max_api)
                if isinstance(update, Update):
                    await self.notify(update, data)
            else:
                raise AttributeError('MaxApi.MaxClient instance not exist')




import abc
from typing import Any, ClassVar
import aiohttp
from pydantic import BaseModel, ConfigDict, Field, PrivateAttr

from .OpcodeEnum import Opcode
from pyromax.mixins import DataBodyMixin, FormDataBodyMixin


class BaseFile(BaseModel):
    model_config = ConfigDict(arbitrary_types_allowed=True, populate_by_name=True)

    max_client: Any | None = Field(default=None, exclude=True)
    url: str | None = None
    uploaded: bool = Field(default=False)

    _data_to_upload: Any = PrivateAttr(default=None)
    _filename: str | None = PrivateAttr(default=None)
    headers: dict | None = Field(default=None, exclude=True)

    async def send_create_request(self):
        return await self.max_client.send_and_receive(opcode=self._opcode, payload={"count": 1})

    @classmethod
    async def upload_new(cls, max_client, data, filename=None) -> 'BaseFile':
        """
        Create object, get URL from server and load bytes
        """

        instance = cls(max_client=max_client, uploaded=False)
        instance._data_to_upload = data
        instance._filename = filename

        await instance.create_cell_for_file()

        await instance.upload_data_to_url()

        return instance

    async def upload_data_to_url(self):
        if not self.url or self._data_to_upload is None:
            return

        try:
            headers = None
            if self.headers:
                headers = self.headers
            elif hasattr(self, '_set_headers') and self._filename:
                self.file_size = len(self._data_to_upload)
                self._set_headers(self.file_size, self._filename)
                headers = self.headers

            async with aiohttp.ClientSession() as session:
                if hasattr(self, 'get_body'):
                    data = self.get_body(self._data_to_upload)
                else:
                    data = self._data_to_upload

                async with session.post(url=self.url, data=data, headers=headers) as response:
                    await self._parse_response(response)

        finally:
            self._data_to_upload = None

    @abc.abstractmethod
    async def create_cell_for_file(self):
        pass

    @abc.abstractmethod
    async def _parse_response(self, response):
        pass


class Photo(FormDataBodyMixin, BaseFile):
    photo_token: str = Field(default='', validation_alias='photoToken')
    url: str | None = Field(default=None, validation_alias='baseUrl')
    photo_id: int | None = Field(default=None, validation_alias='photoId')

    _opcode: ClassVar[int] = Opcode.CREATE_PHOTO.value

    async def create_cell_for_file(self):
        response = await self.send_create_request()
        self.url = response.get('payload', {}).get('url')

    async def _parse_response(self, response):
        photos = (await response.json())['photos']
        photo_id = list(photos.keys())[0]
        self.photo_token = photos[photo_id]['token']
        self.uploaded = True


class Video(DataBodyMixin, BaseFile):
    token: str = Field(default='', validation_alias='token')
    video_id: int = Field(default='', validation_alias='videoId')
    video_type: int | None = Field(default=None, validation_alias='videoType')

    _opcode: ClassVar[int] = Opcode.CREATE_VIDEO.value

    file_size: int | None = Field(default=None, exclude=True)

    async def create_cell_for_file(self):
        if hasattr(self, '_get_payload_info'):
            root = await self._get_payload_info()
        else:
            raise NotImplementedError("Implement _get_payload_info or logic here")
        self.url = root['url']
        self.video_id = root['videoId']
        self.token = root['token']


    async def _parse_response(self, response):
        self.uploaded = True


class File(DataBodyMixin, BaseFile):
    file_id: int = Field(default='', validation_alias='fileId')
    file_name: str = Field(default='', validation_alias='name')
    file_size: int | None = Field(default=None, validation_alias='size')
    file_token: str = Field(default='', validation_alias='token')
    _opcode: ClassVar[int] = Opcode.CREATE_FILE.value


    async def create_cell_for_file(self):
        if hasattr(self, '_get_payload_info'):
            root = await self._get_payload_info()
        else:
            raise NotImplementedError("Implement _get_payload_info")
        self.url = root.get('url')
        self.file_id = root.get('fileId')
        self.file_token = root.get('token')


    async def _parse_response(self, response):
        self.uploaded = True


class Audio(FormDataBodyMixin, BaseFile):
    audio_token: str = Field(default='', validation_alias='token')
    url: str | None = Field(default=None, validation_alias='url')
    audio_id: int | None = Field(default=None, validation_alias='audioId')

    _opcode: ClassVar[int] = Opcode.CREATE_FILE.value

    async def create_cell_for_file(self):
        response = await self.send_create_request()
        self.url = response.get('payload', {}).get('url')

    async def _parse_response(self, response):
        audios = (await response.json())['audios']
        audio_id = list(audios.keys())[0]
        self.photo_token = audios[audio_id]['token']
        self.uploaded = True



from typing import Coroutine

def return_self_after_method(initializer):
    """
    Need use to async initializer in __new__ method
    Args:
          initializer (function): method that will be called, your async initializer
    Returns:
          result of init or self object if initializer returns None

    Examples:
        class ExampleClass1:
            async def your_async_init(self):
                self.test = 'test'

            def __new__(cls, *args, **kwargs):
                self = object.__new__(cls)
                init = return_self_after_method(self.your_async_init)
                return init(self, *args, **kwargs)

            def __repr__(self):
                return 'ExampleClass1'

        class ExampleClass2:
            async def your_async_init(self):
                self.test = 'test'
                return 'init_return_another'

            def __new__(cls, *args, **kwargs):
                self = object.__new__(cls)
                init = return_self_after_method(self.your_async_init)
                return init(self, *args, **kwargs)


        async def main():
             example1 = await ExampleClass1()
             example2 = await ExampleClass2()
             print(example1) # -> 'ExampleClass1'
             print(example2) # -> 'init_return_another'

        if __name__ == '__main__':
            asyncio.run(main())

        Result:
            ExampleClass1
            init_return_another
    """
    async def initializer_wrapper(self, *args, **kwargs) -> Coroutine:
        result = await initializer(self, *args, **kwargs)
        # if result is None:
        #     print('self returned')
        #     return self
        # print('return result')
        # return result
        return result or self
    return initializer_wrapper


from collections.abc import Sequence
from typing import List, Any, TYPE_CHECKING, Awaitable, Callable

from .ObserverPattern import Observer

if TYPE_CHECKING:
    from pyromax.filters import Filter
    from pyromax.api import MaxApi
    from pyromax.types import Update


class Handler(Observer):
    def __init__(self, function, filters: Sequence['Filter'], pattern: None | Callable = None, args: List[Any] = None, from_me=False):
        if args is None:
            args = []
        self.function = function
        self.filters = filters
        self.pattern: Callable[[Any], bool] = pattern
        self.args = args
        self.function: Callable[[Any], Awaitable] = function
        self.pattern = pattern
        self.from_me = from_me


    async def update(self, update, max_api: 'MaxApi', data: dict = None) -> bool:
        if not self.pattern and not self.filters:
            return True
        if update.has_sender_info and (update.sender == max_api.id) != self.from_me:
            return False
        for f in self.filters:
            check = await f(update, max_api=max_api)
            if check:
                if data is not None:
                    data.update({
                        type(check): check
                    })
                return True
        if self.pattern:
            return self.pattern(update)
        return False

    def __repr__(self):
        return f'<{self.__class__.__name__} filters={self.filters}> pattern={self.pattern}> from_me={self.from_me}> args={self.args}>'




from typing import ClassVar

from pydantic import Field, model_validator

from .Update import Update
from .OpcodeEnum import Opcode


class MessageReactionUpdate(Update):

    opcode: ClassVar[int] = Opcode.MESSAGE_REACTION_UPDATE.value
    chat_id: int = Field(alias='chatId')
    message_id: int = Field(alias='messageId')

    type: str = Field(default=None)

    reaction_info: dict = Field(alias='reactionInfo')

    has_sender_info: bool = False

    @model_validator(mode='after')
    def validate_reaction_info(self):
        if self.reaction_info:
            self.type = 'MESSAGE_ADDED_REACTION'
        else:
            self.type = 'MESSAGE_DELETED_REACTION'

        return self


    @classmethod
    def from_update(cls, update: Update) -> 'MessageReactionUpdate':
        return cls(**update.payload, max_api=update.max_api)


    def edit_data(self, data: dict) -> dict:
        data[type(self)] = self
        return data



import random


def get_random_string(length: int, chars: str) -> str:
    """
    Generate a random string of specified length using characters from the provided set.
    
    This function creates a random string by selecting characters randomly
    from the provided character set. Each character in the result is chosen
    independently and uniformly from the chars string.
    
    Args:
        length: The desired length of the random string
        chars: String containing the set of characters to choose from
        
    Returns:
        A random string of the specified length containing only characters from chars
    """
    random_string = ""
    for i in range(length):
        random_string += random.choice(chars)
    return random_string



from pyromax.utils import get_dict_value_by_path

from typing import Any, Callable


class DataBodyMixin:
    def _set_headers(self, file_size, filename):
        self.headers = {
            "Content-Disposition": f"attachment; filename={filename}",
            "Content-Range": f"0-{file_size - 1}/{file_size}",
            "Content-Length": str(file_size),
            "Connection": "keep-alive",
        }

    def get_body(self, data):
        return data

    async def _parse_response(self, response):
        if await response.text() == f"0-{self.file_size - 1}/{self.file_size}":
            self.uploaded = True


    async def _get_payload_info(self):
        response = await self.send_create_request()
        return get_dict_value_by_path('payload info', response)[0]


class FormDataBodyMixin:
    def get_body(self, data):
        return {
            'file': data
        }


from typing import TYPE_CHECKING, Any

if TYPE_CHECKING:
    from pyromax.api import MaxApi


class ReplyMixin:
    max_api: Any
    chat_id: int
    id: str

    async def answer(self, text="", attaches=[]):
        await self.max_api.send_message(chat_id=self.chat_id, text=text, attaches=attaches)


    async def reply(self, text="", attaches=[]):
        await self.max_api.send_message(chat_id=self.chat_id, text=text, attaches=attaches, other_message_elements={
            'link': {
                'type': 'REPLY',
                'messageId': self.id,
            },
        })


# Python imports
from __future__ import annotations
from typing import Any, Optional

# Local imports
from .not_found_flag import NotFoundFlag


def get_dict_value_by_path(path: str, data: dict[str, Any]) -> Optional[dict[str, Any] | NotFoundFlag | str]:
    """
    Navigate through a nested dictionary using a space-separated path and return the value.

    This function traverses a dictionary structure by following a path of keys
    separated by spaces. If all keys in the path exist, the final value is returned.
    If any key in the path is missing, NotFoundFlag is returned.

    Args:
        path: Space-separated string of keys representing the path to navigate
              (e.g., "payload other_directory last_dir")
        data: Dictionary to search in

    Returns:
        The value found at the end of the path, or NotFoundFlag() if the path doesn't exist
        :rtype: dict[str, Any] | NotFoundFlag | None
    """
    layers = path.split()
    current = data
    for layer in layers:
        if not isinstance(current, dict) or layer not in current:
            return NotFoundFlag()
        current = current[layer]
    return current


def has_dict_path(path: str, data: dict[str, Any]) -> bool:
    """
    Check if a path exists in a nested dictionary.

    This function checks whether all keys in a space-separated path exist
    in the provided dictionary structure.

    Args:
        path: Space-separated string of keys representing the path to check
        data: Dictionary to search in

    Returns:
        True if the path exists in the dictionary, False otherwise
    """
    result = get_dict_value_by_path(path, data)
    return not (result == NotFoundFlag)


if __name__ == '__main__':
    test_data = {
        'payload': {
            'other_directory': {
                'last_dir': True
            },
            'another_directory': {
            }
        },
        'idk': 'state'
    }
    result = get_dict_value_by_path('payload other_directory last_dr', test_data)
    print(result)
    if result:
        print('work')


import asyncio
import logging
import pathlib
import string
import time
import uuid
from pathlib import Path
from typing import List, TypeVar, Optional, Union, Dict

import aiofiles
import aiohttp
import qrcode
from aiohttp import ClientTimeout
from websockets.exceptions import ConnectionClosedOK, WebSocketException

from pyromax.api import MaxClient
from pyromax.exceptions import LoggingError, LoggingTimeoutError, SendMessageError, SendMessageFileError, \
    SendMessageNotFoundError
from pyromax.mixins import AsyncInitializerMixin
from pyromax.types import Chat, Opcode, Video, File, Photo, Audio
from pyromax.utils import get_dict_value_by_path
from pyromax.utils import get_random_string
from pyromax.utils.write_token import read_token, write_token

# pprint(dir(Connection))

T = TypeVar('T')
root_dir = Path.cwd()


class MaxApi(AsyncInitializerMixin):
    async def _async_init(self, *args, device_id: str = None, token: str = None,
                          global_context: dict[type[T], T] = None, ping_interval=30, **kwargs):
        if not global_context:
            default_args = {}

        token = await read_token()
        self.__init__(device_id=device_id, token=token, global_context=global_context, ping_interval=ping_interval,
                      *args, **kwargs)

        while True:
            try:
                await self.attach()
                await self.login()
                await self._authorize()
                break
            except WebSocketException:
                await self.detach()
                self.__logger.info('WebSocket cannot init, retrying...')
        self.__logger.info('MaxApi get ready for use')

    async def reload_if_connection_broke(self, dispatcher):
        while True:
            tasks = []
            try:
                tasks.append(asyncio.create_task(dispatcher.start_polling(self)))
                tasks.append(asyncio.create_task(self._keepalive()))

                done, pending = await asyncio.wait(
                    tasks,
                    return_when=asyncio.FIRST_EXCEPTION
                )

                for task in done:
                    if task.exception():
                        raise task.exception()
            except WebSocketException:
                await self.max_client.kill_pending()

                for task in tasks:
                    if not task.done():
                        task.cancel()
                if tasks:
                    await asyncio.gather(*tasks, return_exceptions=True)
                self.__logger.warning('WebSocket connection broke')
                await self.detach()
                await self.attach()
                await self.send_user_agent()
                await self._authorize()

    def __init__(self, device_id: str = None, token: str = None, ping_interval=30, global_context=None, *args,
                 **kwargs) -> None:
        if not global_context:
            global_context = {}
        self.__max_client: MaxClient | None = None
        self._ping_interval: int = ping_interval
        self._track_id: int | None = None
        self.first_name: str | None = None
        self.last_name: str | None = None
        self.name: str | None = None
        self.type: str | None = None
        self.phone: str | None = None
        self.update_time: int | None = None
        self.id: int | None = None
        self.account_status: str | None = None
        self.__logger: logging.Logger = logging.getLogger('MaxApi')
        self.chats: list[Chat] | None = None
        self.start_time = time.time()
        self.__token = token if token else None
        self.device_id = device_id if device_id else self.get_random_device_id()
        self.client_is_login: bool = True if self.__token else False
        self.base_data = {
            MaxApi: self
        }
        self.base_data.update(
            global_context
        )
        self._infinite_recv_task: asyncio.Task | None = None

    # def me(self):
    #     return

    @property
    def max_client(self) -> MaxClient:
        return self.__max_client

    @max_client.setter
    def max_client(self, value: MaxClient) -> None:
        self.__max_client = value

    async def attach(self) -> None:
        if not self.max_client:
            self.max_client = await MaxClient(max_api=self)
            self.start_time = time.time()

    async def detach(self):
        if self.max_client:
            await self.max_client.close_websocket()
            self.max_client = None
        return self.__token, self.device_id, self._ping_interval

    async def _keepalive(self):
        self.__logger.info('Starting keepalive')
        while True:
            self.__logger.debug('Sending keep-alive ping')
            ping = await self.max_client.send_and_receive(opcode=Opcode.PING.value, payload={
                'interactive': False,
            })
            self.__logger.debug('Received keep-alive pong')
            await asyncio.sleep(self._ping_interval)

    @staticmethod
    def get_random_device_id():
        random_string = list(get_random_string(32, string.ascii_lowercase + string.digits))

        for inx in [8, 13, 18, 23]:
            random_string.insert(inx, '-')

        return ''.join(random_string)

    async def send_user_agent(self) -> tuple[int, int, int, str] | None:
        await self.max_client.send_and_receive(opcode=Opcode.SEND_USER_AGENT.value, payload={
            "userAgent": {"deviceType": "WEB", "locale": "ru", "deviceLocale": "ru", "osVersion": "Alpha",
                          "deviceName": "MaxBot",
                          "headerUserAgent": 'Mozilla/5.0 (X11; Linux x86_64; rv:145.0) Gecko/20100101 Firefox/145.0',
                          "appVersion": "25.12.14", "screen": "1440x2560 1.0x", "timezone": "Europe/Moscow"},
            "deviceId": self.device_id}, )

        if self.client_is_login:
            return

        response = await self.max_client.send_and_receive(opcode=Opcode.METADATA_FOR_LOGIN.value, payload='NotSend')

        payload = response[0][0]["payload"]

        polling_interval = payload['pollingInterval'] / 1000
        track_id = payload['trackId']
        expires_at = payload['expiresAt'] / 1000
        url = payload['qrLink']

        return polling_interval, track_id, expires_at, url

    async def login(self):
        async def url_callback_for_login_url(url: str):
            """
            Creating a QR code scanned by max. It is displayed immediately in the console

            Args:
                url - authorization url

            """

            qr = qrcode.QRCode()
            qr.add_data(url)

            qr.make(fit=True)
            qr.print_ascii(invert=True)

        try:
            self.__logger.info('Start Login')

            try:
                metadata = await self.send_user_agent()

                if metadata:
                    self._polling_interval, self._track_id, expires_at, url = metadata
                else:
                    return True
            except KeyError:
                self.__logger.error('Not found attributes in json')
                raise LoggingError('Not found attributes in json')
            self.max_client.websocket.ping_interval = self._polling_interval

            await asyncio.create_task(url_callback_for_login_url(url))

            try:
                async with asyncio.timeout(expires_at - time.time()):
                    while not self.client_is_login:

                        response = await self.max_client.send_and_receive(opcode=Opcode.TRACK_LOGIN.value,
                                                                          payload={"trackId": self._track_id})
                        if response[0][0]["payload"]["status"].get("loginAvailable") is True:
                            self.__logger.info('Login Successful')
                            self.client_is_login = True
                            break
                        await asyncio.sleep(self._polling_interval)
            except asyncio.TimeoutError:
                self.__logger.error('Login Timeout')
                raise LoggingTimeoutError('Login Timeout')
            await self._get_user_data()

            return True
        except ConnectionClosedOK:
            self.__logger.info('Login Failed')
            raise LoggingError('Connection closed')

    async def _parse_user_data(self, response: dict) -> None:
        contact = get_dict_value_by_path('payload profile contact', response)
        names = get_dict_value_by_path('names', contact)
        self.id = get_dict_value_by_path('id', contact)
        self.name = get_dict_value_by_path('name', names)
        self.first_name = get_dict_value_by_path('firstName', names)
        self.last_name = get_dict_value_by_path('lastName', names)
        self.type = get_dict_value_by_path('type', names)
        self.phone = get_dict_value_by_path('phone', contact)
        self.update_time = get_dict_value_by_path('updateTime', contact)

    async def _get_user_data(self) -> None:
        self.__logger.info('Getting User Data...')
        response = await self.max_client.send_and_receive(opcode=Opcode.GET_USER_DATA.value,
                                                          payload={'trackId': self._track_id})
        self.__token = get_dict_value_by_path('payload tokenAttrs LOGIN token', response[0][0])
        await self._parse_user_data(response)

    async def _authorize(self) -> None:
        self.__logger.info('Sending authorize request...')

        response, seq = await self.max_client.send_and_receive(opcode=Opcode.AUTHORIZE.value, payload={
            'interactive': False,
            'token': self.__token,
        })

        await self.max_client.wait_recv(seq - 1, cmd=0)
        token = get_dict_value_by_path('payload token', response[0])
        if token:
            self.__token = str(token)
            await write_token(self.__token)
        await self._parse_user_data(response[0])

        json_chats = response[0]['payload']['chats']

        chats = Chat.from_json(json_chats, self)

        self.chats = chats
        self.__logger.info('Authorized')

    async def get_chat_per_id(self, chat_id: str):
        self.__logger.info('Getting chat info...')
        response = await self.max_client.send_and_receive(opcode=Opcode.GET_CHAT.value, payload={
            'chatIds': [chat_id],
        })

        response = response[0]

        self.__logger.info('Got chat info')

        return Chat(**response['payload']['chats'][0], max_api=self)

    async def send_message(self, chat_id: int, text: str, attaches: List[Video | File | Photo | Audio] = [],
                           other_message_elements: dict = None):
        types_of_attachments = {
            Video: 'VIDEO',
            File: 'FILE',
            Photo: 'PHOTO',
            Audio: 'AUDIO',
        }

        required_params_for_type = {
            'VIDEO': (('videoId', 'video_id'), ('token',)),
            'FILE': (('fileId', 'file_id'),),
            'PHOTO': (('photoToken', 'photo_token'),),
            'AUDIO': (('token', 'audio_token'),),
        }

        loaded_attachments = []
        for attachment in attaches:
            type_of_attachment = types_of_attachments[type(attachment)]

            payload = {
                '_type': type_of_attachment,
            }

            for param in required_params_for_type[type_of_attachment]:
                if len(param) == 1:
                    payload[param[0]] = getattr(attachment, param[0])
                elif len(param) == 2:
                    payload[param[0]] = getattr(attachment, param[1])

            loaded_attachments.append(payload)

        payload = {
            'chatId': chat_id,
            'message': {
                'cid': -round(time.time() * 1000),
                'attaches': loaded_attachments,
            },
        }
        if other_message_elements:
            payload['message'].update(other_message_elements)

        if text:
            payload['message']['text'] = text

        response = await self.max_client.send_and_receive(opcode=Opcode.SEND_MESSAGE.value, payload=payload)
        while error_if_exist := get_dict_value_by_path('payload error', response):
            error_message = get_dict_value_by_path("payload message", response)
            title = get_dict_value_by_path("payload title", response)
            match error_if_exist:
                case 'attachment.not.ready':
                    response = await self.max_client.send_and_receive(opcode=Opcode.SEND_MESSAGE.value, payload=payload)
                    continue
                case 'proto.payload':
                    raise SendMessageFileError(
                        f'''
                        title: {title},
                        error: {error_if_exist},
                        message: {error_message}
                        '''
                    )
                case 'not.found':
                    raise SendMessageNotFoundError(
                        f'''
                        title: {title},
                        error: {error_if_exist},
                        message: {error_message}
                        '''
                    )
                case _:
                    raise SendMessageError(
                        f'''
                        title: {title},
                        error: {error_if_exist},
                        message: {error_message}
                        '''
                    )

    @classmethod
    async def __preparation_data(cls, file: Union[File, Video, Photo, Audio], chat_id: int, message_id: str) -> tuple[
                                                                                                             str, int, dict] | None:
        """
        Prepares data for sending message
        Args:
            file:
            chat_id:
            message_id:

        Returns:
            None if Photo or Audio
            Tuple if File or Video
        """

        if isinstance(file, Photo) or isinstance(file, Audio):
            # Всё есть уже в объекте
            return None


        elif isinstance(file, File):
            path_for_url = "payload url"
            opcode: int = Opcode.DOWNLOAD_FILE.value
            payload: Dict = {
                "fileId": file.file_id,
                "chatId": chat_id,
                "messageId": message_id
            }

        elif isinstance(file, Video):
            path_for_url = "payload MP4_720"

            opcode: int = Opcode.DOWNLOAD_VIDEO.value
            payload = {
                "videoId": file.video_id,
                "chatId": chat_id,
                "messageId": message_id
            }

        else:
            raise TypeError("Unknown media type")

        final_result = (path_for_url, opcode, payload)

        return final_result

    @classmethod
    async def __download_file(cls, url: str, path_for_upload: Optional[str | Path] = None) -> Union[
        bytes, str]:
        """

        Args:
            url: url for download

            path_for_upload: path for save file if None, then it is saved in a byte stream

        Returns:

        """

        chunks: list[bytes] = []
        headers = {
            "User-Agent": "Mozilla/5.0 (X11; Linux x86_64; rv:145.0) Gecko/20100101 Firefox/145.0",
            "Accept": "*/*",
            "Referer": "https://ok.ru/",
        }

        cookies = {
            "tstc": "p"
        }
        # Для видео передаю заголовки и куки, т.к видео с серваков одноклассников
        async with aiohttp.ClientSession(headers=headers, cookies=cookies, timeout=ClientTimeout(total=60)) as session:
            async with session.get(url) as response:

                if response.status != 200:
                    raise RuntimeError("Download failed")

                if path_for_upload:
                    pth_upload = pathlib.Path(path_for_upload)
                    filename = response.headers.get("X-File-Name")

                    if filename is None or pathlib.Path.exists(pth_upload / filename):
                        filename_first = f"{uuid.uuid4()}"
                    else:
                        filename_first = filename
                    # Файл уже получает расширение без нативного расширения. Костыль, но можно изменить
                    async with aiofiles.open(pth_upload / filename_first, 'wb') as f:
                        async for chunk in response.content.iter_chunked(8192):
                            await f.write(chunk)
                    return filename_first
                else:
                    async for chunk in response.content.iter_chunked(8192):
                        chunks.append(chunk)
                    return b"".join(chunks)

    async def download_file(self, file: Union[Photo, Audio, File, Video], chat_id: int, message_id: str,
                            path_for_download: Optional[str | Path] = None) -> Union[bytes, str]:

        """

        Args:
            file: file object
            chat_id: chat id for download
            message_id: message id for download
            path_for_download: path for download file
        Returns:

        """
        data_for_get_url = await self.__preparation_data(file, chat_id, message_id)
        if data_for_get_url:
            path_for_url, opcode, payload = data_for_get_url[0], data_for_get_url[1], data_for_get_url[2]

            response = await self.max_client.send_and_receive(opcode=opcode, payload=payload)

            url_for_download = get_dict_value_by_path(
                path=path_for_url,
                data=response[0][0],
            )
        else:
            url_for_download = file.url
        result = await self.__download_file(url=url_for_download,
                                            path_for_upload=path_for_download)
        return result


async def main():
    logging.basicConfig(level=logging.INFO)


if __name__ == '__main__':
    asyncio.run(main())



from pyromax.exceptions.LoggingError import LoggingError, LoggingTimeoutError
from pyromax.exceptions.BaseMaxApiException import BaseMaxApiException
from pyromax.exceptions.SendMessageError import SendMessageError, SendMessageFileError, SendMessageNotFoundError
from pyromax.exceptions.AnnotationHandlerError import AnnotationHandlerError

__all__ = ['LoggingError', 'LoggingTimeoutError', 'BaseMaxApiException', 'SendMessageError', 'SendMessageFileError', 'SendMessageNotFoundError', 'AnnotationHandlerError']


import asyncio
import json
from pathlib import Path
import aiofiles

ROOT_DIR = Path().resolve()
JSON_FILE = ROOT_DIR / "tokens.json"

async def write_token(token: str):
    """Write or overwriting token from json file."""

    data = {"max_token": token}
    async with aiofiles.open(JSON_FILE, mode="w") as f:
        await f.write(json.dumps(data, indent=4))

async def read_token() -> str | None:
    """Read token from json file."""
    if not JSON_FILE.exists():
        return None
    async with aiofiles.open(JSON_FILE, mode="r") as f:
        content = await f.read()
        if not content.strip():
            return None
        try:
            data = json.loads(content)
            return data.get("max_token")
        except json.JSONDecodeError:
            return None

async def main():
    await write_token("test")

    token = await read_token()
    print(token)

if __name__ == "__main__":
    asyncio.run(main())


from .BaseMaxApiException import BaseMaxApiException

class LoggingError(BaseMaxApiException):
    """Exception raised when a logging error occurs."""


class LoggingTimeoutError(BaseMaxApiException):
    """Exception raised when a logging timeout occurs."""



from abc import ABC, abstractmethod
from collections.abc import Awaitable, Callable
from typing import TYPE_CHECKING, Any

# COPY-PASTED FROM AIOGRAM

class Filter(ABC):
    """
    If you want to register own filters like builtin filters you will need to write subclass
    of this class with overriding the :code:`__call__`
    method and adding filter attributes.
    """

    if TYPE_CHECKING:
        # This checking type-hint is needed because mypy checks validity of overrides and raises:
        # error: Signature of "__call__" incompatible with supertype "BaseFilter"  [override]
        # https://mypy.readthedocs.io/en/latest/error_code_list.html#check-validity-of-overrides-override
        __call__: Callable[..., Awaitable[bool | dict[str, Any]]]
    else:  # pragma: no cover

        @abstractmethod
        async def __call__(self, *args: Any, **kwargs: Any) -> bool | dict[str, Any]:
            """
            This method should be overridden.

            Accepts incoming event and should return boolean or dict.

            :return: :class:`bool` or :class:`dict[str, Any]`
            """




from .BaseMaxApiException import BaseMaxApiException

class SendMessageError(BaseMaxApiException):
    """Exception raised when a sending message fails."""



class SendMessageFileError(SendMessageError):
    """Exception raised when a sending message fails because the message with the attached file contained a text or/and other attachments."""


class SendMessageNotFoundError(SendMessageError):
    """Exception raised when a sending message fails because the chat not found"""


from pydantic import BaseModel, Field, AliasPath

from typing import TYPE_CHECKING, Any


if TYPE_CHECKING:
    from ..api.MaxApi import MaxApi


class Update(BaseModel):
    max_api: Any = Field(exclude=True)
    opcode: int
    type: str = Field(exclude=True, default=None)
    payload: dict | None = Field(default=None, exclude=True)





from collections.abc import Generator
from typing import List, Optional

from pyromax.api.observer.event import MaxEventObserver
from pyromax.api.observer import ObserverPattern
from pyromax.api import MaxApi
from pyromax.types import (
    Message,
    MessageReactionUpdate,
    Opcode, Update
)


class Router(ObserverPattern.Subject):
    def __init__(self):
        self.sub_routers: list[Router] = []
        self._parent_router = None

        self.message = MaxEventObserver(self, 'USER', opcode=Opcode.PUSH_NOTIFICATION.value, type_of_update=Message)
        self.edited_message = MaxEventObserver(self, 'EDITED_MESSAGE', opcode=Opcode.PUSH_NOTIFICATION.value, type_of_update=Message)
        self.reply_to_message = MaxEventObserver(self, 'REPLY', opcode=Opcode.PUSH_NOTIFICATION.value, type_of_update=Message)
        self.message_added_reaction = MaxEventObserver(self, 'MESSAGE_ADDED_REACTION', opcode=Opcode.MESSAGE_REACTION_UPDATE.value, type_of_update=MessageReactionUpdate)
        self.message_deleted_reaction = MaxEventObserver(self, 'MESSAGE_DELETED_REACTION', opcode=Opcode.MESSAGE_REACTION_UPDATE.value, type_of_update=MessageReactionUpdate)
        self.events = {
            'MESSAGE': self.message,
            'EDITED_MESSAGE': self.edited_message,
            'REPLY_TO_MESSAGE': self.reply_to_message,
            'MESSAGE_ADDED_REACTION': self.message_added_reaction,
            'MESSAGE_DELETED_REACTION': self.message_deleted_reaction,
        }



    @property
    def chain_head(self) -> Generator['Router', None, None]:
        router: Router | None = self
        print(self.sub_routers)
        while router:
            yield router
            router = router.parent_router

    @property
    def chain_tail(self) -> Generator['Router', None, None]:
        yield self
        for router in self.sub_routers:
            yield from router.chain_tail

    @property
    def parent_router(self) -> Optional['Router']:
        return self._parent_router

    @parent_router.setter
    def parent_router(self, router: 'Router') -> None:
        """
        Internal property setter of parent router fot this router.
        Do not use this method in own code.
        All routers should be included via `include_router` method.

        Self- and circular- referencing are not allowed here

        :param router:
        """
        if not isinstance(router, Router):
            msg = f"router should be instance of Router not {type(router).__name__!r}"
            raise ValueError(msg)
        if self._parent_router:
            msg = f"Router is already attached to {self._parent_router!r}"
            raise RuntimeError(msg)
        if self == router:
            msg = "Self-referencing routers is not allowed"
            raise RuntimeError(msg)

        parent: Router | None = router
        while parent is not None:
            if parent == self:
                msg = "Circular referencing of Router is not allowed"
                raise RuntimeError(msg)

            parent = parent.parent_router

        self._parent_router = router
        router.sub_routers.append(self)

    def include_routers(self, *routers: 'Router') -> None:
        """
        Attach multiple routers.

        :param routers:
        :return:
        """
        if not routers:
            msg = "At least one router must be provided"
            raise ValueError(msg)
        for router in routers:
            self.include_router(router)

    def include_router(self, router: 'Router') -> 'Router':
        """
        Attach another router.

        :param router:
        :return:
        """
        if not isinstance(router, Router):
            msg = f"router should be instance of Router not {type(router).__class__.__name__}"
            raise ValueError(msg)
        router.parent_router = self
        return router

    async def notify(self, update: Update, data = None) -> None:
        if data is None:
            raise ValueError("data cannot be None")

        for event in self.events.values():
            if event.opcode != update.opcode:
                continue
            handler, data = await event.update(update, data[MaxApi], data=data)
            if handler and event.event_name == update.type:
                args = [data[arg] for arg in handler.args if arg in data]
                return await handler.function(*args)

        for router in self.sub_routers:
            return await router.notify(update=update, data=data)
            # router.notify(update=update, data=data)


    # def include_router(self, router):
    #     for name, event in self.events.items():
    #         event: MaxEventObserver
    #         event.include_event(router.events.pop(name))
    #
    #
    # def include_routers(self, routers):
    #     for router in routers:
    #         self.include_router(router)



from pyromax.utils.get_random_string import get_random_string
from pyromax.utils.get_dict_value_by_path import has_dict_path, get_dict_value_by_path
from pyromax.utils.not_found_flag import NotFoundFlag
from pyromax.utils.decorator_applier import apply_decorator_to_method
from pyromax.utils.return_self import return_self_after_method


__all__ = [
    "get_random_string",
    "has_dict_path",
    "get_dict_value_by_path",
    "NotFoundFlag",
    "apply_decorator_to_method",
    'return_self_after_method',
]


from __future__ import annotations


class NotFoundFlag:
    """
    Sentinel value used to indicate that a value was not found.
    
    This class serves as a special marker object that can be returned
    when a search or lookup operation fails. It is always falsy and
    compares equal to other NotFoundFlag instances or the NotFoundFlag type.
    """

    def __eq__(self, other: NotFoundFlag) -> bool:
        """
        Compare this instance with another object.
        
        Returns True if other is a NotFoundFlag instance or the NotFoundFlag type,
        False otherwise. This allows checking for "not found" using either
        `result == NotFoundFlag` or `result == NotFoundFlag()`.
        
        Args:
            other: Object to compare with
            
        Returns:
            True if other is NotFoundFlag or NotFoundFlag type, False otherwise
        """
        if type(other) == NotFoundFlag or (type(self) == type(other()) if type(other) == type else None):
            return True
        return False

    def __bool__(self) -> bool:
        """
        Return the boolean value of this instance.
        
        Always returns False, making NotFoundFlag instances falsy.
        This allows using NotFoundFlag in boolean contexts like:
        `if not result:` or `if result: ... else: ...`
        
        Returns:
            Always False
        """
        return False



from pyromax.mixins.FileMixins import DataBodyMixin, FormDataBodyMixin
from pyromax.mixins.AsyncInitializer import AsyncInitializerMixin
from pyromax.mixins.ReplyMixin import ReplyMixin

__all__ = ['DataBodyMixin', 'FormDataBodyMixin', 'AsyncInitializerMixin', 'ReplyMixin']



import asyncio
import json
import logging
import re
import string
from typing import TYPE_CHECKING

from websockets import connect
from websockets.typing import Origin

from pyromax.mixins import AsyncInitializerMixin
from pyromax.utils import get_random_string

if TYPE_CHECKING:
    pass


class MaxClient(AsyncInitializerMixin):
    async def _async_init(self, max_api):
        self.__init__(max_api)
        self._recv_task = asyncio.create_task(self.infinite_recv())
        await self._init_websocket()

    def __init__(self, max_api):
        self.max_api = max_api
        self.__logger: logging.Logger = logging.getLogger('MaxClient')
        self.__polling_interval: int = 0
        self.websocket: connect = None
        self._inited: asyncio.Event = asyncio.Event()
        self.__counter: int = 0
        self.sec_websocket_key: str = get_random_string(23, string.ascii_uppercase + string.digits)
        self.__pending_requests: dict[str, list[asyncio.Future]] = {}
        self.__message_buffer: dict[str, list[dict]] = {}
        self.__running_lock = asyncio.Lock()
        self.__recv_event: asyncio.Event = asyncio.Event()
        # self.inited = asyncio.Event()
        self._recv_exception: Exception | None = None
        self._recv_task: asyncio.Task | None = None
        self._websocket_inited = asyncio.Event()

    @property
    def counter(self):
        return self.__counter

    def counter_increment(self):
        self.__counter += 1
        return self.__counter

    @property
    def polling_interval(self):
        return self.__polling_interval

    @polling_interval.setter
    def polling_interval(self, value):
        self.__polling_interval = value

    async def _init_websocket(self):
        if not self.websocket:
            self.__logger.info('Initializing Max WebSocket Client')
            self.websocket = await connect("wss://ws-api.oneme.ru/websocket",
                                           origin=Origin('https://web.max.ru'),
                                           user_agent_header='Mozilla/5.0 (X11; Linux x86_64; rv:145.0) Gecko/20100101 Firefox/145.0',
                                           ping_interval=self.__polling_interval)
        self._websocket_inited.set()

    async def close_websocket(self):
        if self.websocket:
            await self.websocket.close()
            self.websocket.ws_client = None
            self.max_api = None
            self.__logger.info('WebSocket Closed')
        self._websocket_inited.clear()

    def receive_is_locked(self) -> bool:
        return self.__running_lock.locked()

    async def kill_pending(self):
        for requests in self.__pending_requests.values():
            for request in requests:
                request.cancel()

            if requests:
                await asyncio.gather(*requests, return_exceptions=True)

    async def infinite_recv(self):

        self.__logger.info('Starting infinite receiving')
        try:
            await self._websocket_inited.wait()
            while True:
                response = json.loads(await self.websocket.recv())

                opcode = response['opcode']
                cmd = response['cmd']
                seq = response['seq']
                request_signature = f'{cmd}{seq}{opcode}'
                handled = False

                for pattern in self.__pending_requests.copy():
                    if re.search(pattern, request_signature):
                        for request in self.__pending_requests[pattern]:
                            if not request.done():
                                request.set_result([response])
                                handled = True
                        del self.__pending_requests[pattern]
                if not handled:
                    if request_signature in self.__message_buffer:
                        self.__message_buffer[request_signature].append(response)
                    else:
                        self.__message_buffer[request_signature] = [response]

        finally:
            pass

    async def send_message(self, message: dict, send_count: int = 1) -> None | int:
        for _ in range(send_count):
            await self.websocket.send(json.dumps(message))
        return message['seq']

    async def wait_recv(self, seq: int = None, cmd: int = 1, opcode: int | None = None) -> list[dict]:
        """
        Package 01292 is static, and code 292 is sending a banner.
        It stops all authorization, so it is immediately deleted and is an empty package.
        Args:
            seq:
            cmd:
            opcode:
        """
        loop = asyncio.get_running_loop()
        response_future = loop.create_future()

        opcode_pattern = str(opcode) if opcode is not None else r'[0-9]+'
        seq_pattern = str(seq) if seq is not None else r'[0-9]+'
        pattern = f'{cmd}{seq_pattern}{opcode_pattern}'

        for key in list(self.__message_buffer.keys()):
            if key == "01292":
                self.__message_buffer.pop(key)
                return []

            if re.search(pattern, key) or key.endswith("288") or key.endswith("289"):
                messages = self.__message_buffer.pop(key)
                response_future.set_result(messages)
                return response_future.result()

        if pattern not in self.__pending_requests:
            self.__pending_requests[pattern] = [response_future]
        else:
            self.__pending_requests[pattern].append(response_future)

        if not self._recv_task or self._recv_task.done():
            if self._recv_task and self._recv_task.exception():
                raise self._recv_task.exception()
            raise ConnectionError("Websocket reader is not running")

        try:
            result = await asyncio.wait_for(response_future, timeout=60)
            return result
        except asyncio.TimeoutError:
            return []
        finally:
            if response_future.cancelled():
                for requests in self.__pending_requests.values():
                    if response_future in requests:
                        requests.remove(response_future)

    async def send_and_receive(self, ver: int = 11, opcode: int = 1, cmd: int = 0, payload: dict | str = None):
        request = {
            'ver': ver,
            'opcode': opcode,
            'cmd': cmd,
            'seq': self.counter_increment(),
        }

        if payload != 'NotSend':
            request['payload'] = payload
        seq = await self.send_message(request)

        response = await self.wait_recv(seq=seq, cmd=int(not cmd), opcode=opcode)

        await asyncio.sleep(self.__polling_interval)
        return response, seq


async def main():
    max_client = await MaxClient.create_client()


if __name__ == '__main__':
    logging.basicConfig(level=logging.INFO)

    asyncio.run(main())



# Python imports
from __future__ import annotations
from typing import Any, Callable


async def apply_decorator_to_method(
    obj: Any,
    method_name: str,
    decorator: Callable[[Any, Callable[..., Any]], Any]
) -> None:
    """
    Apply a decorator function to a method of an object and replace the original method.
    
    This function retrieves the original method from the object, applies the decorator
    to it (which may be async), and then replaces the original method with the decorated version.
    
    Args:
        obj: The object whose method should be decorated
        method_name: Name of the method to decorate
        decorator: Async callable that takes (obj, original_method) and returns
                   a decorated version of the method
                   
    Example:
        >>> async def my_decorator(obj, method):
        ...     async def wrapper(*args, **kwargs):
        ...         # Do something before
        ...         result = await method(*args, **kwargs)
        ...         # Do something after
        ...         return result
        ...     return wrapper
        >>> await apply_decorator_to_method(client, 'send_message', my_decorator)
    """
    original_method = getattr(o=obj, name=method_name)
    decorated_method = await decorator(obj, original_method)
    setattr(obj=obj, name=method_name, value=decorated_method)



from abc import ABC, abstractmethod


from pyromax.types import Update
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from .. import MaxApi


class Observer(ABC):
    @abstractmethod
    async def update(self, update: Update, max_api: 'MaxApi') -> bool:
        ...


class Subject(ABC):


    async def attach(self, handler: Observer) -> None:
        if handler not in self._handlers:
            self._handlers.append(handler)


    async def detach(self, handler: Observer) -> None:
        if handler in self._handlers:
            self._handlers.remove(handler)


    @abstractmethod
    async def notify(self, update: Update) -> None:
        ...





from typing import TYPE_CHECKING, Union
import inspect


from pyromax.exceptions import AnnotationHandlerError
from pyromax.api.observer import ObserverPattern

from pyromax.api.observer import Handler
from pyromax.api.observer import Router

if TYPE_CHECKING:
    from pyromax.filters import Filter
    from pyromax.api import MaxApi
    from pyromax.types import Update

class MaxEventObserver(ObserverPattern.Observer):
    """
        Event observer for Max events

        Here you can register handler with filter.
        This observer will stop event propagation when first handler is pass.
        """

    def __init__(self, router: 'Router', event_name: str, opcode: int, type_of_update) -> None:
        self.type_of_update = type_of_update
        self.opcode = opcode
        self.router: Router = router
        self.event_name: str = event_name
        self.handlers: list['Handler'] = []

        # self.middleware = MiddlewareManager()
        # self.outer_middleware = MiddlewareManager()

        # Re-used filters check method from already implemented handler object
        # with dummy callback which never will be used
        # self._handler = Handler(pattern=lambda: True, filters=[])

    async def update(self, update, max_api: 'MaxApi', data: dict = None) -> tuple['Handler',  dict] | tuple[bool, dict]:
        if data is None:
            raise ValueError('data cannot be None')
        for handler in self.handlers:
            if await handler.update(update, max_api, data=data):
                return handler, data
        return False, data


    def include_event(self, event: 'MaxEventObserver'):
        self.handlers += event.handlers


    def include_events(self, events):
        for event in events:
            self.include_event(event)


    def __call__(self, *filters: 'Filter', pattern=None, from_me: bool = False):
        def decorator(func):
            signature = inspect.signature(func)
            args = [param.annotation for param in signature.parameters.values()]
            if inspect._empty in args:
                raise AnnotationHandlerError('Need annotation all params in handler')
            handler = Handler(func, filters=filters, pattern=pattern, args=args, from_me=from_me)
            self.handlers.append(handler)
        return decorator





from enum import Enum

class Opcode(Enum):
    PING = 1
    SEND_USER_AGENT = 6
    AUTHORIZE = 19
    CHAT_INFO = 48
    GET_CHAT_MESSAGES_PER_CHUNK = 49
    SEND_MESSAGE = 64
    GET_MESSAGE = 71
    CREATE_PHOTO = 80
    CREATE_VIDEO = 82
    DOWNLOAD_VIDEO = 83
    CREATE_FILE = 87
    DOWNLOAD_FILE = 88
    PUSH_NOTIFICATION = 128
    MESSAGE_REACTION_UPDATE = 156
    METADATA_FOR_LOGIN = 288
    TRACK_LOGIN = 289
    GET_USER_DATA = 291



from .File import Video, Photo, File, Audio
from pyromax.mixins.ReplyMixin import ReplyMixin
from .Update import Update
from .OpcodeEnum import Opcode

from pydantic import ConfigDict, field_validator, Field, AliasPath, PrivateAttr, model_validator

from typing import TYPE_CHECKING, ClassVar

if TYPE_CHECKING:
    from ..api import MaxApi


class Message(Update, ReplyMixin):
    id: str
    text: str
    sender: int
    time: int
    type: str
    attaches: list
    chat_id: int = Field(alias='chatId')
    has_sender_info: bool = True

    type: str = Field(alias='type', validation_alias=AliasPath('link', 'type'), default='USER', exclude=True) # Can be "REPLY", "EDITED" or "USER"
    status: str | None = Field(default=None, exclude=True)

    opcode: ClassVar[int] = Opcode.PUSH_NOTIFICATION.value


    @field_validator('attaches', mode='after')
    @classmethod
    def attaches_to_model(cls, attaches: list) -> list[Video | Photo | File | Audio]:
        types_of_attaches = {
            'VIDEO': Video,
            'PHOTO': Photo,
            'FILE': File,
            'AUDIO': Audio,
        }

        attaches_valid = []
        for attach in attaches:
            if attach['_type'] in types_of_attaches:
                attaches_valid.append(types_of_attaches[attach['_type']](**attach))
            else:
                attaches_valid.append(attach)
        return attaches_valid


    @model_validator(mode='after')
    def type_of_message(self):
        if self.status == 'EDITED_MESSAGE':
            self.type = self.status
        return self


    @classmethod
    def from_update(cls, update: Update) -> 'Message':
        self = cls(**update.model_dump(), **update.payload, **update.payload['message'], max_api=update.max_api)
        return self


    def edit_data(self, data: dict) -> dict:
        data[type(self)] = self
        return data


    def __repr__(self):
        details = []
        for key, value in self.__dict__.items():
            details.append(f"{key}={value}")
        return '\n'.join(details)


from typing import Coroutine

from pyromax.utils.return_self import return_self_after_method

import abc


class AsyncInitializerMixin(abc.ABC):
    """
    Added async initializer in class, for begin you need inherit from AsyncInitializerMixin and create a method with
    _async_init name.
    You can nothing return in _async_init, and in default it be return self
    or
    You can return anything except None
    Coroutines can be awaited in this init(in example: await asyncio.sleep(1))
    Example:
        class NeedAsyncInit(AsyncInitializerMixin):
            async def _init(self):
                await asyncio.sleep(1)
                self.test = 'test'

            def __repr__(self):
                return 'NeedAsyncInit'

        class AnotherNeedAsyncInit(AsyncInitializerMixin):
            async def _init(self):
                await asyncio.sleep(1)
                self.test = 'test'
                return 'another return'

        async def main():
            example1 = NeedAsyncInit()
            example2 = AnotherNeedAsyncInit()
            print(example1)
            print(example2)


        if __name__ == '__main__':
            asyncio.run(main())

        CONSOLE:
            NeedAsyncInit
            another return
    """

    def __new__(cls, *args, **kwargs) -> Coroutine:
        self = object.__new__(cls)
        init = return_self_after_method(cls._async_init)
        return init(self, *args, **kwargs)

    def __await__(self):
        """just blank"""
        ...
        return self

    @abc.abstractmethod
    async def _async_init(self, *args, **kwargs):
        ...


from pyromax.types.Message import Message
from pyromax.types.Chat import Chat
from pyromax.types.Update import Update
from pyromax.types.OpcodeEnum import Opcode
from pyromax.types.File import Video, Photo, File, Audio
from pyromax.types.MessageReactionUpdate import MessageReactionUpdate

__all__ = ['Message', 'Chat', 'Update', 'Opcode', 'Audio', 'File', 'Video', 'Photo', 'MessageReactionUpdate']